@@ -1,58 +1,62 @@
 from __future__ import annotations
 
 import math
-import random
 from contextlib import suppress
 from copy import copy
 from typing import Dict, List, Any, Optional, Union, Callable
 
 import glm
 from OpenGL.GL import glReadPixels
+from OpenGL.raw.GL.ARB.vertex_shader import GL_FLOAT
 from OpenGL.raw.GL.VERSION.GL_1_0 import glEnable, GL_TEXTURE_2D, GL_DEPTH_TEST, glClearColor, glClear, \
-    GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT, GL_RGBA, GL_BLEND, glBlendFunc, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, \
-    glDisable, GL_CULL_FACE, GL_CW, GL_BACK, glCullFace, GL_FRONT_AND_BACK
+    GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT, GL_BLEND, glBlendFunc, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, \
+    glDisable, GL_CULL_FACE, GL_BACK, glCullFace, GL_DEPTH_COMPONENT
 from OpenGL.raw.GL.VERSION.GL_1_2 import GL_UNSIGNED_INT_8_8_8_8, GL_BGRA
 from glm import mat4, vec3, quat, vec4
+from pykotor.common.geometry import Vector2, Vector3
+
 from pykotor.common.misc import CaseInsensitiveDict
 from pykotor.common.module import Module
 from pykotor.common.stream import BinaryReader
 from pykotor.extract.file import ResourceIdentifier
 from pykotor.extract.installation import Installation, SearchLocation
-from pykotor.resource.formats.lyt import LYT
+from pykotor.resource.formats.lyt import LYT, LYTRoom
 from pykotor.resource.formats.tpc import TPC
-from pykotor.resource.formats.twoda import read_2da
+from pykotor.resource.formats.twoda import read_2da, TwoDA
 from pykotor.resource.generics.git import GIT, GITPlaceable, GITCreature, GITDoor, GITTrigger, GITEncounter, \
     GITWaypoint, GITSound, GITStore, GITCamera, GITInstance
 from pykotor.resource.type import ResourceType
 
 from pykotor.gl.shader import Shader, KOTOR_VSHADER, KOTOR_FSHADER, Texture, PICKER_FSHADER, PICKER_VSHADER, \
     PLAIN_VSHADER, PLAIN_FSHADER
-from pykotor.gl.modelreader import gl_load_mdl, gl_load_stitched_model
-from pykotor.gl.model import Model, Cube, Boundary, Empty
-from pykotor.gl.models.predefined import STORE_MDL_DATA, STORE_MDX_DATA, WAYPOINT_MDL_DATA, WAYPOINT_MDX_DATA, \
+from pykotor.gl.models.read_mdl import gl_load_stitched_model
+from pykotor.gl.models.mdl import Model, Cube, Boundary, Empty
+from pykotor.gl.models.predefined_mdl import STORE_MDL_DATA, STORE_MDX_DATA, WAYPOINT_MDL_DATA, WAYPOINT_MDX_DATA, \
     SOUND_MDL_DATA, SOUND_MDX_DATA, CAMERA_MDL_DATA, CAMERA_MDX_DATA, TRIGGER_MDL_DATA, TRIGGER_MDX_DATA, \
-    ENCOUNTER_MDL_DATA, ENCOUNTER_MDX_DATA, ENTRY_MDL_DATA, ENTRY_MDX_DATA, EMPTY_MDL_DATA, EMPTY_MDX_DATA
+    ENCOUNTER_MDL_DATA, ENCOUNTER_MDX_DATA, ENTRY_MDL_DATA, ENTRY_MDX_DATA, EMPTY_MDL_DATA, EMPTY_MDX_DATA, \
+    CURSOR_MDX_DATA, CURSOR_MDL_DATA, UNKNOWN_MDL_DATA, UNKNOWN_MDX_DATA
 
 SEARCH_ORDER_2DA = [SearchLocation.OVERRIDE, SearchLocation.CHITIN]
 SEARCH_ORDER = [SearchLocation.CUSTOM_MODULES, SearchLocation.OVERRIDE, SearchLocation.CHITIN]
 
 
 class Scene:
-    SPECIAL_MODELS = ["waypoint", "store", "sound", "camera", "trigger", "encounter"]
+    SPECIAL_MODELS = ["waypoint", "store", "sound", "camera", "trigger", "encounter", "unknown"]
 
-    def __init__(self, installation: Installation, *, module: Optional[Module] = None):
+    def __init__(self, *, installation: Optional[Installation] = None, module: Optional[Module] = None):
         glEnable(GL_TEXTURE_2D)
         glEnable(GL_DEPTH_TEST)
         glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
         glCullFace(GL_BACK)
 
-        self.installation: Installation = installation
+        self.installation: Optional[Installation] = installation
         self.textures: CaseInsensitiveDict[Texture] = CaseInsensitiveDict()
         self.models: CaseInsensitiveDict[Model] = CaseInsensitiveDict()
         self.objects: List[RenderObject] = []
         self.selection: List[RenderObject] = []
         self.module: Optional[Module] = module
         self.camera: Camera = Camera()
+        self.cursor: RenderObject = RenderObject("cursor")
 
         self.textures["NULL"] = Texture.from_color()
 
@@ -67,10 +71,12 @@ class Scene:
 
         self.jumpToEntryLocation()
 
-        self.table_doors = read_2da(installation.resource("genericdoors", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
-        self.table_placeables = read_2da(installation.resource("placeables", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
-        self.table_creatures = read_2da(installation.resource("appearance", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
-        self.table_heads = read_2da(installation.resource("heads", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
+        self.table_doors = TwoDA()
+        self.table_placeables = TwoDA()
+        self.table_creatures = TwoDA()
+        self.table_heads = TwoDA()
+        if installation is not None:
+            self.setInstallation(installation)
 
         self.hide_creatures: bool = False
         self.hide_placeables: bool = False
@@ -85,8 +91,19 @@ class Scene:
         self.hide_trigger_boundaries: bool = True
         self.hide_encounter_boundaries: bool = True
         self.backface_culling: bool = True
+        self.use_lightmap: bool = True
+        self.show_cursor: bool = True
+
+    def setInstallation(self, installation: Installation) -> None:
+        self.table_doors = read_2da(installation.resource("genericdoors", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
+        self.table_placeables = read_2da(installation.resource("placeables", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
+        self.table_creatures = read_2da(installation.resource("appearance", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
+        self.table_heads = read_2da(installation.resource("heads", ResourceType.TwoDA, SEARCH_ORDER_2DA).data)
 
     def buildCache(self, clearCache: bool = False) -> None:
+        if self.module is None:
+            return
+
         if clearCache:
             self.objects = {}
 
@@ -123,7 +140,7 @@ class Scene:
         for room in self.layout.rooms:
             if room not in self.objects:
                 position = vec3(room.position.x, room.position.y, room.position.z)
-                self.objects[room] = RenderObject(room.model, position)
+                self.objects[room] = RenderObject(room.model, position, data=room)
 
         for door in self.git.doors:
             if door not in self.objects:
@@ -132,7 +149,7 @@ class Scene:
                     model_name = self.table_doors.get_row(utd.appearance_id).get_string("modelname")
                 except Exception:
                     # If failed to load creature models, use an empty model instead
-                    model_name = "empty"
+                    model_name = "unknown"
 
                 self.objects[door] = RenderObject(model_name, vec3(), vec3(), data=door)
 
@@ -146,7 +163,7 @@ class Scene:
                     model_name = self.table_placeables.get_row(utp.appearance_id).get_string("modelname")
                 except Exception:
                     # If failed to load creature models, use an empty model instead
-                    model_name = "empty"
+                    model_name = "unknown"
 
                 self.objects[placeable] = RenderObject(model_name, vec3(), vec3(), data=placeable)
 
@@ -169,7 +186,7 @@ class Scene:
                         obj.children.append(head_obj)
                 except Exception:
                     # If failed to load creature models, use an empty model instead
-                    obj = RenderObject("empty", vec3(), vec3(), data=creature)
+                    obj = RenderObject("unknown", vec3(), vec3(), data=creature)
 
                 self.objects[creature] = obj
 
@@ -227,9 +244,9 @@ class Scene:
                 obj = RenderObject("camera", vec3(), vec3(), data=camera)
                 self.objects[camera] = obj
 
-            self.objects[camera].set_position(camera.position.x, camera.position.y, camera.position.z)
+            self.objects[camera].set_position(camera.position.x, camera.position.y, camera.position.z+camera.height)
             euler = glm.eulerAngles(quat(camera.orientation.w, camera.orientation.x, camera.orientation.y, camera.orientation.z))
-            self.objects[camera].set_rotation(euler.y, euler.z-math.pi/2, -euler.x+math.pi/2)
+            self.objects[camera].set_rotation(euler.y, euler.z-math.pi/2+math.radians(camera.pitch), -euler.x+math.pi/2)
 
         # Detect if GIT still exists; if they do not then remove them from the render list
         for obj in copy(self.objects):
@@ -254,16 +271,20 @@ class Scene:
 
     def render(self) -> None:
         self.buildCache()
+
         glClearColor(0.5, 0.5, 1, 1.0)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
 
         if self.backface_culling:
             glEnable(GL_CULL_FACE)
+        else:
+            glDisable(GL_CULL_FACE)
 
         glDisable(GL_BLEND)
         self.shader.use()
         self.shader.set_matrix4("view", self.camera.view())
         self.shader.set_matrix4("projection", self.camera.projection())
+        self.shader.set_bool("enableLightmap", self.use_lightmap)
         group1 = [obj for obj in self.objects.values() if obj.model not in self.SPECIAL_MODELS]
         for obj in group1:
             self._render_object(self.shader, obj, mat4())
@@ -285,7 +306,7 @@ class Scene:
 
         # Draw boundary for selected objects
         glDisable(GL_CULL_FACE)
-        self.plain_shader.set_vector4("color", vec4(0.0, 1.0, 0.0, 0.4))
+        self.plain_shader.set_vector4("color", vec4(0.0, 1.0, 0.0, 0.8))
         for obj in self.selection:
             obj.boundary(self).draw(self.plain_shader, obj.transform())
 
@@ -297,6 +318,10 @@ class Scene:
         for obj in [obj for obj in self.objects.values() if obj.model == "trigger" and not self.hide_trigger_boundaries]:
             obj.boundary(self).draw(self.plain_shader, obj.transform())
 
+        if self.show_cursor:
+            self.plain_shader.set_vector4("color", vec4(1.0, 0.0, 0.0, 0.4))
+            self._render_object(self.plain_shader, self.cursor, mat4())
+
     def _render_object(self, shader: Shader, obj: RenderObject, transform: mat4) -> None:
         if isinstance(obj.data, GITCreature) and self.hide_creatures:
             return
@@ -328,6 +353,8 @@ class Scene:
 
         if self.backface_culling:
             glEnable(GL_CULL_FACE)
+        else:
+            glDisable(GL_CULL_FACE)
 
         self.picker_shader.use()
         self.picker_shader.set_matrix4("view", self.camera.view())
@@ -344,6 +371,25 @@ class Scene:
             self._picker_render_object(obj, mat4())
 
     def _picker_render_object(self, obj: RenderObject, transform: mat4) -> None:
+        if isinstance(obj.data, GITCreature) and self.hide_creatures:
+            return
+        if isinstance(obj.data, GITPlaceable) and self.hide_placeables:
+            return
+        if isinstance(obj.data, GITDoor) and self.hide_doors:
+            return
+        if isinstance(obj.data, GITTrigger) and self.hide_triggers:
+            return
+        if isinstance(obj.data, GITEncounter) and self.hide_encounters:
+            return
+        if isinstance(obj.data, GITWaypoint) and self.hide_waypoints:
+            return
+        if isinstance(obj.data, GITSound) and self.hide_sounds:
+            return
+        if isinstance(obj.data, GITStore) and self.hide_sounds:
+            return
+        if isinstance(obj.data, GITCamera) and self.hide_cameras:
+            return
+
         model = self.model(obj.model)
         model.draw(self.picker_shader, transform * obj.transform())
         for child in obj.children:
@@ -359,6 +405,7 @@ class Scene:
         if clear_existing:
             self.selection.clear()
 
+        self.buildCache()
         if isinstance(target, GITInstance):
             for obj in self.objects.values():
                 if obj.data is target:
@@ -367,6 +414,27 @@ class Scene:
 
         self.selection.append(target)
 
+    def screenToWorld(self, x: int, y: int) -> Vector3:
+        glClearColor(0.5, 0.5, 1, 1.0)
+        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
+
+        if self.backface_culling:
+            glEnable(GL_CULL_FACE)
+        else:
+            glDisable(GL_CULL_FACE)
+
+        glDisable(GL_BLEND)
+        self.shader.use()
+        self.shader.set_matrix4("view", self.camera.view())
+        self.shader.set_matrix4("projection", self.camera.projection())
+        group1 = [obj for obj in self.objects.values() if isinstance(obj.data, LYTRoom)]
+        for obj in group1:
+            self._render_object(self.shader, obj, mat4())
+
+        zpos = glReadPixels(x, self.camera.height-y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT)[0][0]
+        cursor = glm.unProject(vec3(x, self.camera.height-y, zpos), self.camera.view(), self.camera.projection(), vec4(0, 0, self.camera.width, self.camera.height))
+        return Vector3(cursor.x, cursor.y, cursor.z)
+
     def texture(self, name: str) -> Texture:
         if name not in self.textures:
             try:
@@ -383,6 +451,9 @@ class Scene:
         return self.textures[name]
 
     def model(self, name: str) -> Model:
+        mdl_data = EMPTY_MDL_DATA
+        mdx_data = EMPTY_MDX_DATA
+
         if name not in self.models:
             if name == "waypoint":
                 mdl_data = WAYPOINT_MDL_DATA
@@ -408,15 +479,18 @@ class Scene:
             elif name == "empty":
                 mdl_data = EMPTY_MDL_DATA
                 mdx_data = EMPTY_MDX_DATA
-            else:
+            elif name == "cursor":
+                mdl_data = CURSOR_MDL_DATA
+                mdx_data = CURSOR_MDX_DATA
+            elif name == "unknown":
+                mdl_data = UNKNOWN_MDL_DATA
+                mdx_data = UNKNOWN_MDX_DATA
+            elif self.installation is not None:
                 mdl_search = self.installation.resource(name, ResourceType.MDL, SEARCH_ORDER, capsules=self.module.capsules())
                 mdx_search = self.installation.resource(name, ResourceType.MDX, SEARCH_ORDER, capsules=self.module.capsules())
                 if mdl_search and mdx_search:
                     mdl_data = mdl_search.data
                     mdx_data = mdx_search.data
-                else:
-                    mdl_data = EMPTY_MDL_DATA
-                    mdx_data = EMPTY_MDX_DATA
 
             # model = gl_load_mdl(self, BinaryReader.from_bytes(mdl_data, 12), BinaryReader.from_bytes(mdx_data))
             try:
@@ -528,11 +602,12 @@ class Camera:
         self.x: float = 40.0
         self.y: float = 130.0
         self.z: float = 0.5
+        self.width: int = 1920
+        self.height: int = 1080
         self.pitch: float = math.pi / 2
         self.yaw: float = 0.0
         self.distance: float = 10.0
         self.fov: float = 90.0
-        self.aspect: float = 16 / 9
 
     def view(self) -> mat4:
         up = vec3(0, 0, 1)
@@ -550,7 +625,7 @@ class Camera:
         return view
 
     def projection(self) -> mat4:
-        return glm.perspective(self.fov, self.aspect, 0.1, 5000)
+        return glm.perspective(self.fov, self.width/self.height, 0.1, 5000)
 
     def translate(self, translation: vec3) -> None:
         self.x += translation.x
@@ -589,4 +664,4 @@ class Camera:
         x += math.cos(self.yaw) * math.cos(self.pitch - math.pi / 2) * self.distance
         y += math.sin(self.yaw) * math.cos(self.pitch - math.pi / 2) * self.distance
         z += math.sin(self.pitch - math.pi / 2) * self.distance
-        return vec3(x, y, z)
+        return vec3(x, y, z)
\ No newline at end of file
